package writecache

import (
	"context"
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/internal/logs"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/shard/mode"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	"go.etcd.io/bbolt"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
)

type setModePrm struct {
	ignoreErrors bool
	shrink       bool
}

// SetMode sets write-cache mode of operation.
// When shard is put in read-only mode all objects in memory are flushed to disk
// and all background jobs are suspended.
func (c *cache) SetMode(m mode.Mode) error {
	ctx, span := tracing.StartSpanFromContext(context.TODO(), "writecache.SetMode",
		trace.WithAttributes(
			attribute.String("mode", m.String()),
		))
	defer span.End()

	c.modeMtx.Lock()
	defer c.modeMtx.Unlock()

	err := c.setMode(ctx, m, setModePrm{ignoreErrors: true})
	if err == nil {
		c.metrics.SetMode(mode.ConvertToComponentModeDegraded(m))
	}
	return err
}

// setMode applies new mode. Must be called with cache.modeMtx lock taken.
func (c *cache) setMode(ctx context.Context, m mode.Mode, prm setModePrm) error {
	var err error
	turnOffMeta := m.NoMetabase()

	if turnOffMeta && !c.mode.NoMetabase() {
		err = c.flush(ctx, prm.ignoreErrors)
		if err != nil {
			return err
		}
	}

	if err := c.closeDB(prm.shrink); err != nil {
		return err
	}

	// Suspend producers to ensure there are channel send operations in fly.
	// flushCh is populated by `flush` with `modeMtx` taken, thus waiting until it is empty
	// guarantees that there are no in-fly operations.
	for len(c.flushCh) != 0 {
		c.log.Info(logs.WritecacheWaitingForChannelsToFlush)
		time.Sleep(time.Second)
	}

	if turnOffMeta {
		c.mode = m
		return nil
	}

	if err = c.openStore(mode.ConvertToComponentModeDegraded(m)); err != nil {
		return err
	}

	c.mode = m
	return nil
}

func (c *cache) closeDB(shrink bool) error {
	if c.db == nil {
		return nil
	}
	if !shrink {
		if err := c.db.Close(); err != nil {
			return fmt.Errorf("can't close write-cache database: %w", err)
		}
		return nil
	}

	var empty bool
	err := c.db.View(func(tx *bbolt.Tx) error {
		b := tx.Bucket(defaultBucket)
		empty = b == nil || b.Stats().KeyN == 0
		return nil
	})
	if err != nil && !errors.Is(err, bbolt.ErrDatabaseNotOpen) {
		return fmt.Errorf("failed to check DB items: %w", err)
	}
	if err := c.db.Close(); err != nil {
		return fmt.Errorf("can't close write-cache database: %w", err)
	}
	if empty {
		err := os.Remove(filepath.Join(c.path, dbName))
		if err != nil && !os.IsNotExist(err) {
			return fmt.Errorf("failed to remove DB file: %w", err)
		}
	} else {
		c.log.Info(logs.WritecacheShrinkSkippedNotEmpty)
	}
	return nil
}

// readOnly returns true if current mode is read-only.
// `c.modeMtx` must be taken.
func (c *cache) readOnly() bool {
	return c.mode.ReadOnly()
}

// noMetabase returns true if c is operating without the metabase.
// `c.modeMtx` must be taken.
func (c *cache) noMetabase() bool {
	return c.mode.NoMetabase()
}
